"""Translate coordinates made for transformer neural networks back into ints/floats."""

import numpy as np


def right_pack_array(in_array):
    """
    Packs a boolean array into uint8s. Unlike NumPy's packbits, this function pads to the left.

    >>> right_pack_array(np.array([0,0,1,1,0,1,0,0,1,1,0,1,0,0,1,0,1,0,1,0,1,0,1]))
    array([ 26, 105,  85], dtype=uint8)

    :param in_array:
    :return:
    """
    ext_in_array = np.zeros((8 - (len(in_array)) % 8,), dtype=np.bool)
    padded_in_array = np.concatenate((ext_in_array, in_array))
    packed_in = np.packbits(padded_in_array)
    return packed_in


def packed_unit8_to_int(in_array):
    """
    Packs a uint8 array into a single int.

    >>> packed_unit8_to_int(np.array([ 26, 105,  85], dtype=np.uint8))
    1730901

    :param in_array:
    :return:
    """
    num_i = int(0)
    for i, a in enumerate(reversed(in_array)):
        num_i += int(a) << (8 * i)
    return num_i


def right_pack_bool_to_int(in_array):
    """
    Convert an array of booleans to a packed array of unit8s.

    This is one part of getting numbers back out of pcoord format. Another is undoing the ith_middle change.

    >>> right_pack_bool_to_int(np.array([0,0,0,1,1,0,1,0,0,1,1,0,1,0,0,1,0,1,0,1,0,1,0,1]))
    1730901

    """
    arr1 = right_pack_array(in_array)
    arr2 = packed_unit8_to_int(arr1)
    return arr2



"""Translate int/float coordinates to formats more suitable for transformer neural networks."""

import math as m

import numpy as np

from pnums.ith_middle import ith_middle, mask_ith_middle
from typing import Dict, Any


def _int_to_packed_uint8(in_int):
    """
    Convert an int to a unit array.

    >>> _int_to_packed_uint8(16)
    array([16], dtype=uint8)

    """
    in_bytes = int(in_int).to_bytes(
        length=int(m.ceil(m.log2(in_int + 1) / 8.0)), byteorder="big", signed=False
    )
    in_array = np.frombuffer(in_bytes, dtype=np.uint8)
    return in_array


def int_to_bool_array(in_int, length):
    """
    Convert an int to a bool array.

    >>> int_to_bool_array(16, 32)
    array([0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
           0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 1., 0., 0., 0., 0.])

    """
    byte_arr = np.unpackbits(_int_to_packed_uint8(in_int))
    if len(byte_arr) < length:
        pad_l = np.zeros((length - len(byte_arr),))
        byte_arr = np.concatenate((pad_l, byte_arr))
    return byte_arr


def inversion_double(in_array):
    """
    Get the input boolean array along with its element-wise logical not beside it. For error correction.

    >>> inversion_double(np.array([1,0,1,1,1,0,0,1], dtype=np.bool))
    array([[ True, False,  True,  True,  True, False, False,  True],
           [False,  True, False, False, False,  True,  True, False]])
    """
    return np.stack((in_array, np.logical_not(in_array)))


_i_extend_dict: Dict[Any, Any] = {}  # optimization dictionary


def _i_extend(in_array, ext, sort=True):
    """
    Create `ext`*2 duplicates and shifts the duplicates so each number has a unique representation.

    Error correction could fetch back two numbers if they were recorded on the same array.
    Useful for neural algorithms that would guess multiple locations represented by these coordiantes.

    >>> _i_extend(np.array([ 1, 0, 1, 1, 1, 0, 0, 1], dtype=np.bool), 8).astype(np.uint8)
    array([[[1, 0, 1, 1, 1, 0, 0, 1],
            [0, 1, 0, 0, 0, 1, 1, 0]],
    <BLANKLINE>
           [[1, 1, 1, 1, 1, 0, 0, 1],
            [0, 0, 0, 0, 0, 1, 1, 0]],
    <BLANKLINE>
           [[1, 1, 0, 1, 1, 0, 0, 1],
            [0, 0, 1, 0, 0, 1, 1, 0]],
    <BLANKLINE>
           [[1, 1, 0, 1, 1, 0, 0, 1],
            [0, 0, 1, 0, 0, 1, 1, 0]],
    <BLANKLINE>
           [[1, 1, 0, 0, 1, 0, 1, 1],
            [0, 0, 1, 1, 0, 1, 0, 0]],
    <BLANKLINE>
           [[0, 1, 0, 0, 1, 0, 1, 1],
            [1, 0, 1, 1, 0, 1, 0, 0]],
    <BLANKLINE>
           [[0, 1, 0, 0, 1, 1, 1, 1],
            [1, 0, 1, 1, 0, 0, 0, 0]],
    <BLANKLINE>
           [[0, 1, 0, 0, 0, 1, 1, 1],
            [1, 0, 1, 1, 1, 0, 0, 0]]], dtype=uint8)

    """
    base_array = inversion_double(in_array)
    base_array = np.expand_dims(base_array, axis=0)

    if sort:
        if ext in _i_extend_dict.keys():
            i_list = _i_extend_dict[ext]
        else:
            i_list = [
                ith
                for _, ith in sorted(
                    zip(
                        [ith_middle(i) for i in range(ext - 1)],
                        [j for j in range(ext - 1)],
                    )
                )
            ]
            _i_extend_dict[ext] = i_list
    else:
        i_list = range(ext - 1)

    for i in i_list:
        next_array = mask_ith_middle(in_array, i)
        double_next_array = inversion_double(next_array)
        double_next_array = np.expand_dims(double_next_array, axis=0)
        base_array = np.concatenate((base_array, double_next_array), axis=0)
    return base_array


def int_to_1d(in_int, max_bits, new_dim_length):
    """Convert an integer to a 1D format with error correction for transformers."""
    bool_array = int_to_bool_array(in_int, max_bits)
    extended_bool_array = _i_extend(bool_array, new_dim_length)
    return extended_bool_array


def ints_to_2d(
    in_int_1, max_bits_1, new_dim_length_1, in_int_2, max_bits_2, new_dim_length_2
):
    """Convert two integers representing x,y coordinates to a correlated output for 2D transformers."""
    bool_ento_1 = int_to_1d(in_int_1, max_bits_1, new_dim_length_1).astype(np.bool)
    bool_ento_2 = int_to_1d(in_int_2, max_bits_2, new_dim_length_2).astype(np.bool)
    positive_array = np.zeros(
        (new_dim_length_1, new_dim_length_2, max(max_bits_1, max_bits_2)), dtype=np.bool
    )
    for i in range(bool_ento_1.shape[-1]):
        region_1 = bool_ento_1[:, 0, i, np.newaxis]
        region_2 = bool_ento_2[np.newaxis, :, 0, i]
        square_region = np.matmul(region_1, region_2)
        positive_array[:, :, i] = square_region
    return positive_array



"""Ith middle functions for masking."""

import math as m
import numpy as np


def ith_middle(i, min_value=0.0, max_value=1.0):
    """
    Return the ith middle.

    >>> for i in range(10):
    ...   ith_middle(i)
    0.5
    0.25
    0.75
    0.125
    0.375
    0.625
    0.875
    0.0625
    0.1875
    0.3125

    """
    tree_level = m.floor(m.log2(i + 1))
    leaves_up_to_level = 2 ** tree_level
    id_in_level = i - leaves_up_to_level + 1
    value_for_level = (id_in_level + 0.5) / leaves_up_to_level
    middle_value = min_value + value_for_level * (max_value - min_value)
    return middle_value


def ith_binary_mask(i):
    """
    Return the ith binary mask for novel binary number representation.

    For one bit in a binary integer, take the bit and all bits below it until the length of the array is met.
    Add the resulting integer and this array together. The topmost bit will be one in the in a region around the
    ith middle. (Allow for overflow.)

    >>> for i in range(10):
    ...   ith_binary_mask(i)
    array([False,  True])
    array([False, False,  True])
    array([False,  True,  True])
    array([False, False, False,  True])
    array([False, False,  True,  True])
    array([False,  True, False,  True])
    array([False,  True,  True,  True])
    array([False, False, False, False,  True])
    array([False, False, False,  True,  True])
    array([False, False,  True, False,  True])

    """
    assert i < 256
    tree_level = m.floor(m.log2(i + 1))
    bit_mask = np.zeros((2 + tree_level), dtype=bool)
    bit_mask[-1] = 1
    leaves_up_to_level = 2 ** tree_level
    id_in_level = i - leaves_up_to_level + 1
    upper_bool = np.unpackbits(np.array([id_in_level], dtype=np.uint8))
    bit_mask[:-1] = upper_bool[-len(bit_mask) + 1 :]
    return bit_mask


def mask_ith_middle(in_array, i):
    """
    Transform a boolean array so its int form, a, is a*ith_middle(i) away.

    >>> mask_ith_middle(np.array([1,0,1,1,1,0,0,1], dtype=np.uint8), 0)
    array([1, 1, 0, 0, 1, 0, 1, 1], dtype=uint8)

    """
    i_array = ith_binary_mask(i)
    packed_i = np.packbits(i_array)
    ext_array = np.zeros((7,))
    padded_in_array = np.concatenate((in_array, ext_array)).astype(np.bool)
    return_array = np.zeros_like(in_array)
    for i in range(len(in_array)):
        pad_num = padded_in_array[i : i + 8]
        packed_num = np.packbits(pad_num)
        added_num = np.add(packed_num, packed_i)
        bool_added_num = np.unpackbits(added_num)
        return_array[i] = bool_added_num[0]
    return return_array


def unmask_ith_middle(in_array, i):
    """
    Transform a boolean array so its int form, a, is a*-ith_middle(i) away.

    #>>> unmask_ith_middle(np.array([1, 1, 0, 0, 1, 0, 1, 1], dtype=np.uint8), 0)
    array([1, 0, 1, 1, 1, 0, 0, 1], dtype=uint8)

    """
    raise NotImplementedError("Haven't figured this out yet.")



"""Work in progress."""
import math
import warnings
import numpy as np
from numbers import Real


def _binlist_to_int(binlist):
    """
    Convert a binary list to int.

    Do not use. This should be replaces with the similar function in xcoords_to_int.
    """
    # from: https://stackoverflow.com/a/12461400
    out = 0
    for bit in binlist:
        out = (out << 1) | int(bit)
    return out


class NeuralFloat(Real):
    """
    Converts A floating point number to a binary array that's closer to what the brain uses, centered at zero.

    The benefit of using this is that similar numbers have a lot of similar bits on (or "activated neurons"), so
    networks that connect to this number to learn will be able to learn be able to learn ranges. Or areas for the 2D
    version of this.

    >>> a = NeuralFloat(3.14, min_val=.001, max_val=100)
    >>> print(a)
    [[0. 0. 0. 0. 0. 1. 1. 0. 0. 1. 0. 0. 0. 1. 1. 1. 1. 0. 1. 0. 1. 1. 1. 0.
      0. 0. 0. 1. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 1.]
     [0. 0. 0. 0. 0. 0. 0. 1. 1. 0. 1. 1. 1. 0. 0. 0. 0. 1. 0. 1. 0. 0. 0. 1.
      1. 1. 1. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0.]]
    >>> print(float(a))
    3.1399998664855957

    >>> b = NeuralFloat(3.14*3, min_val=.001, max_val=100)
    >>> print(b)
    [[0. 0. 0. 1. 0. 0. 1. 0. 1. 1. 0. 1. 0. 1. 1. 1. 0. 0. 0. 0. 1. 0. 1. 0.
      0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 1.]
     [0. 0. 0. 0. 1. 1. 0. 1. 0. 0. 1. 0. 1. 0. 0. 0. 1. 1. 1. 1. 0. 1. 0. 1.
      1. 1. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0.]]
    >>> print(float(b))
    9.419998168945312
    """

    def __init__(self, num=0.0, min_val=2 ** -126, max_val=2 ** 127, mantissa_bits=23):
        """Create a 32-bit neural float based on the input number."""
        self.float_val = num
        self.down_bits = int(math.ceil(math.log2(1 / min_val)))
        if self.down_bits >= 127:
            warnings.warn(
                "Allowing minimum possible float values will require many bits. Please limit if possible."
            )
        self.up_bits = int(math.ceil(math.log2(max_val)))
        if self.up_bits >= 126:
            warnings.warn(
                "Allowing maximum possible float valuse will require many bits. Please limit if possible."
            )
        self.mantissa_bits = int(mantissa_bits)
        self.bits = self.mantissa_bits + self.up_bits + self.down_bits

        self._array = self.convert_float_to_nb32(num)

    def convert_float_to_nb32(self, float_in):
        """Convert an ieee float to 32-bit neural binary float."""
        z = np.zeros((2, self.bits))
        mantissa, exp = math.frexp(float_in)
        if mantissa < 0:
            sign = 0
        else:
            sign = 1
        if exp > self.up_bits:
            exp = self.up_bits
            mantissa_array = [1 for _ in range(self.mantissa_bits)]
        elif exp < -self.down_bits:
            exp = -self.down_bits
            mantissa_array = [0 for _ in range(self.mantissa_bits)]
        else:
            mantissa_array = [
                int(x) for x in bin(int(abs(mantissa * 2 ** self.mantissa_bits)))[2:]
            ]
        active_bits = np.array(mantissa_array, dtype=np.bool)
        z[0, self.up_bits - exp : self.up_bits - exp + self.mantissa_bits] = active_bits
        z[1, self.up_bits - exp : self.up_bits - exp + self.mantissa_bits] = np.invert(
            active_bits
        )
        z[0, -1] = sign
        z[1, -1] = not sign
        return z

    def __float__(self):
        start = min(np.nonzero(self._array == 1)[1])
        mantissa_array = self._array[0, start : start + self.mantissa_bits]
        mantissa = _binlist_to_int(mantissa_array) / (2 ** self.mantissa_bits)
        exp = self.up_bits - start
        msign = self._array[1, -1]
        psign = self._array[0, -1]
        out = (msign * -1 + psign * 1) * mantissa * 2.0 ** exp
        return out

    def indexes(self):
        """Get the nonzero indexes of the float."""
        return np.nonzero(np.ravel(self._array))

    def __str__(self):
        return str(self._array)

    def __repr__(self):
        return str(self)

    def __add__(self, other):
        pass

    def __radd__(self, other):
        pass

    def __neg__(self):
        pass

    def __pos__(self):
        pass

    def __mul__(self, other):
        pass

    def __rmul__(self, other):
        pass

    def __truediv__(self, other):
        pass

    def __rtruediv__(self, other):
        pass

    def __pow__(self, exponent):
        pass

    def __rpow__(self, base):
        pass

    def __abs__(self):
        pass

    def __trunc__(self):
        pass

    def __floor__(self):
        pass

    def __ceil__(self):
        pass

    def __round__(self, ndigits=None):
        pass

    def __floordiv__(self, other):
        pass

    def __rfloordiv__(self, other):
        pass

    def __mod__(self, other):
        pass

    def __rmod__(self, other):
        pass

    def __lt__(self, other):
        pass

    def __le__(self, other):
        pass

    def __eq__(self, other):
        pass


class NeuralRobustFloat(Real):
    """
    Uses a list of masks to convert a NeuralFloat to a set of new representations, which can all be unmasked.

    The benefit of using this is that if the binary values are combined later, multiple numbers can be taken out
    instead of one

    >>> a = NeuralFloat(3.14, min_val=.001, max_val=100)
    >>> print(a)
    [[0. 0. 0. 0. 0. 1. 1. 0. 0. 1. 0. 0. 0. 1. 1. 1. 1. 0. 1. 0. 1. 1. 1. 0.
      0. 0. 0. 1. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 1.]
     [0. 0. 0. 0. 0. 0. 0. 1. 1. 0. 1. 1. 1. 0. 0. 0. 0. 1. 0. 1. 0. 0. 0. 1.
      1. 1. 1. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0.]]
    >>> print(float(a))
    3.1399998664855957

    >>> b = NeuralFloat(3.14*3, min_val=.001, max_val=100)
    >>> print(b)
    [[0. 0. 0. 1. 0. 0. 1. 0. 1. 1. 0. 1. 0. 1. 1. 1. 0. 0. 0. 0. 1. 0. 1. 0.
      0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 1.]
     [0. 0. 0. 0. 1. 1. 0. 1. 0. 0. 1. 0. 1. 0. 0. 0. 1. 1. 1. 1. 0. 1. 0. 1.
      1. 1. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0.]]
    >>> print(float(b))
    9.419998168945312
    """

    pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pnums']

package_data = \
{'': ['*']}

install_requires = \
['numpy==1.16.1', 'scipy>=1.4.1,<2.0.0']

setup_kwargs = {
    'name': 'pnums',
    'version': '0.0.2',
    'description': 'A library that encodes coordinates so neural networks can use them better.',
    'long_description': 'PCoords\n=======\nNeural Coordinates\n\nInstallation\n------------\nWill be pip installable on first release.',
    'author': 'SimLeek',
    'author_email': 'simulator.leek@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/simleek/pnums',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



"""Probabilistic n-dimensional binary integer format."""

from numbers import Real

import numpy as np
from scipy.ndimage import convolve  # type: ignore

from pnums.int_to_xcoords import int_to_bool_array, inversion_double
from pnums.xcoords_to_int import right_pack_bool_to_int


def shift(arr, num, fill_value=np.nan):
    """Shift data in array arr num positions left or right."""
    result = np.empty_like(arr)
    if num > 0:
        result[..., :num] = fill_value
        result[..., num:] = arr[..., :-num]
    elif num < 0:
        result[..., num:] = fill_value
        result[..., :num] = arr[..., -num:]
    else:
        result[:] = arr
    return result


def layer_xor(a: np.ndarray, b: np.ndarray) -> np.ndarray:
    """Run an xor operation between two layers of neurons."""
    o = convolve(a, b, mode="wrap")
    s = np.sum(o)
    if s != 0:
        return o / s
    else:
        return layer_zero(o)


def layer_zero(a: np.ndarray):
    """Return the PInt representation of a 0 bit for a layer."""
    o = np.zeros_like(a)
    o[tuple(1 for _ in range(o.ndim))] = 1
    return o


def layer_one(a: np.ndarray):
    """Return the PInt representation of a 1 bit for a layer."""
    o = np.zeros_like(a)
    o[tuple(0 for _ in range(o.ndim))] = 1
    return o


def layer_and(a: np.ndarray, b: np.ndarray) -> np.ndarray:
    """Run an and operation between two layers of neurons."""
    if a.ndim > 1:
        a_max = np.argwhere(np.max(a) == a)
        b_max = np.argwhere(np.max(b) == b)
        c_sum = np.sum((1 - a_max) * (1 - b_max))
        a_sum = np.sum(a_max)
        b_sum = np.sum(b_max)
        if a_sum > b_sum:
            o = a * np.sum(b)
        elif a_sum == b_sum or c_sum == 0:
            o = a * b
        else:
            o = b * np.sum(a)
    else:
        o = a * b
    s = np.sum(o)
    if s != 0:
        return o / s
    else:
        return layer_zero(o)


def layer_or(a: np.ndarray, b: np.ndarray) -> np.ndarray:
    """Run an or operation between two layers of neurons."""
    o = convolve(a, b, mode="constant", cval=0)
    s = np.sum(o)
    if s != 0:
        return o / s
    else:
        return layer_zero(o)


class PInt(Real):
    """Probabilistic representation of any number of correlated integers."""

    def __init__(self, *args, bits=32, confidence=1.0, dtype=np.float32):
        """
        Probabilistic representation of any number of correlated integers.

        :param args: a list of numbers or numpy array may be inputted
        :param bits: number of bits used to represent each number
        :param confidence: probability of this set of correlated numbers
        :param dtype: internal storage type
        """
        if isinstance(args[0], np.ndarray):
            self.tensor = args[0]
            self.ndim = self.tensor.ndim - 1
            self.bits = self.tensor.shape[-1]
        else:
            if dtype not in {
                np.float,
                np.float16,
                np.float32,
                np.float64,
                np.double,
                np.half,
            }:
                raise NotImplementedError(
                    "PInt currently only stores probabilities as values between 0 and 1. "
                    "Non-real types like int or complex will need different methods."
                )
            self.ndim = len(args)
            self.tensor = np.ones([2] * self.ndim + [bits], dtype=dtype)
            self.bits = bits

            for d in range(self.ndim):
                in_val = inversion_double(int_to_bool_array(args[d], bits)).astype(
                    np.float
                )
                if in_val.shape[-1] < bits:
                    last_slice = in_val.shape[-1]
                else:
                    last_slice = bits
                self.tensor[
                    (slice(2),) * d
                    + (0,)
                    + (slice(2),) * (self.ndim - d - 1)
                    + (slice(last_slice),)
                ] *= in_val[0, -bits:]
                self.tensor[
                    (slice(2),) * d
                    + (1,)
                    + (slice(2),) * (self.ndim - d - 1)
                    + (slice(last_slice),)
                ] *= in_val[1, -bits:]
                anti_tensor = (1 - self.tensor) * (
                    (1.0 - confidence) / (2 ** self.ndim - 1)
                )
            self.tensor *= confidence
            self.tensor += anti_tensor
        self.confidence = confidence

    def normalize(self, confidence=1.0):
        """Make each bit/layer add up to the same max probability."""
        new_tensor = np.zeros_like(self.tensor)
        for b in reversed(range(self.bits)):
            if np.sum(self.tensor[..., b]) != 0:
                new_tensor[..., b] = (
                    self.tensor[..., b] / np.sum(self.tensor[..., b]) * confidence
                )
        return PInt(new_tensor)

    def quantize(self):
        """Make each internal number one or zero. Useful for transforming back to normal numbers for computers."""
        new_tensor = np.zeros_like(self.tensor)
        for b in reversed(range(self.bits)):
            if np.sum(self.tensor[..., b]) != 0:
                max_indices = np.argwhere(
                    self.tensor[..., b] == np.amax(self.tensor[..., b])
                )
                ind = max_indices[0]
                if len(max_indices) == 1:
                    new_tensor[(*ind, b)] = 1
        return PInt(new_tensor)

    def coalesce(self, value=1.0):
        """
        Interpolate between normalized and quantized PInts.

        Useful for keeping a stable representation, like Kalman filtering.
        """
        multiplier = self.quantize().tensor * value
        new_tensor = np.zeros_like(self.tensor)
        norm = self.normalize()
        for b in reversed(range(self.bits)):
            new_tensor[..., b] = multiplier[..., b] + (
                (1.0 - value) * norm.tensor[..., b]
            )
        return PInt(new_tensor)

    def _handle_other_types(self, other):
        if isinstance(other, PInt):
            other_tensor = other.tensor
        elif isinstance(other, np.ndarray):
            other_tensor = other
        else:
            raise NotImplementedError(f"{type(other)} type currently not implemented.")
        return other_tensor

    def __add__(self, other):
        """Full adder implemented with probabilities."""
        other_tensor = self._handle_other_types(other)
        self_confidence = self.confidence
        other_confidence = other.confidence
        half_sum_prob = np.zeros_like(self.tensor)
        sum_prob = np.zeros_like(self.tensor)
        carry_prob = np.zeros_like(self.tensor)
        carry_prob[tuple(1 for _ in range(self.ndim)) + (self.bits - 1,)] = 1
        for b in reversed(range(self.bits)):
            half_sum_prob[..., b] = layer_xor(self.tensor[..., b], other_tensor[..., b])
            sum_prob[..., b] = layer_xor(half_sum_prob[..., b], carry_prob[..., b])
            if b > 0:
                carry_prob[..., b - 1] = layer_or(
                    layer_and(self.tensor[..., b], other_tensor[..., b]),
                    layer_and(half_sum_prob[..., b], carry_prob[..., b]),
                )

        return PInt(sum_prob).normalize(self_confidence + other_confidence)

    def __sub__(self, other):
        other_tensor = self._handle_other_types(other)
        return self.__add__(1 - other_tensor)

    def asfloat(self):
        """Return the floating point representation of all corellated numbers held in this PInt."""
        quant = self.normalize().quantize().tensor.astype(np.bool)
        if self.ndim == 1:
            val = right_pack_bool_to_int(quant[0, ...])
            return float(val)
        else:
            vals = []
            for n in range(self.ndim):
                pos_quant = quant[
                    tuple(slice(2) for _ in range(n))
                    + (0,)
                    + tuple(slice(2) for _ in range(n + 1, self.ndim))
                ]
                sum_quant = np.sum(
                    pos_quant, axis=tuple(i for i in range(self.ndim - 1))
                )
                val = right_pack_bool_to_int(sum_quant)
                vals.append(val)
            return tuple(vals)

    def __float__(self):
        f = self.asfloat()
        if isinstance(f, float):
            return f
        elif isinstance(f, tuple):
            val = 0
            for n, i in enumerate(f):
                val += n * (2 ** self.bits) ** i
            return val

    def __trunc__(self):
        f = self.asfloat()
        if isinstance(f, float):
            return int(f)
        elif isinstance(f, tuple):
            val = 0
            for n, i in enumerate(f):
                val += n * (2 ** self.bits) ** i
            return int(val)

    def __floor__(self):
        return self.__trunc__()

    def __ceil__(self):
        return self.__trunc__()

    def __round__(self, ndigits=None):
        return self.__trunc__()

    def __floordiv__(self, other):
        pass

    def __rfloordiv__(self, other):
        pass

    def __mod__(self, other):
        pass

    def __rmod__(self, other):
        pass

    def __lt__(self, other):
        pass

    def __le__(self, other):
        pass

    def __radd__(self, other):
        pass

    def __neg__(self):
        pass

    def __pos__(self):
        pass

    def __mul__(self, other):
        pass

    def __rmul__(self, other):
        pass

    def __truediv__(self, other):
        pass

    def __rtruediv__(self, other):
        pass

    def __pow__(self, exponent):
        pass

    def __rpow__(self, base):
        pass

    def __abs__(self):
        pass

    def __eq__(self, other):
        pass
